"""
General library containing metrics and helper functions.

.. moduleauthor:: Leiden University Medical Center <humgen@lumc.nl>
.. moduleauthor:: Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
.. moduleauthor:: Martijn Vermaat <m.vermaat.hg@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future import standard_library

import numpy as np

with standard_library.hooks():
    from collections import Counter


def distribution(vector):
    """
    Calculate the distribution of the values in a vector.

    :arg vector: A vector.
    :type vector: iterable(int)

    :return: A list of `(value, count)` pairs.
    :rtype: list(int, int)
    """
    # Todo: I'm not sure this should be in this module.
    return sorted(Counter(vector).items())


def vector_length(vector):
    """
    Calculate the Euclidean length of a vector.

    :arg array_like vector: A vector.

    :return: The length of `vector`.
    :rtype: float
    """
    # Note: This seems faster than `numpy.linalg.norm`.
    return np.sqrt(np.dot(vector, vector))


def get_scale(left, right):
    """
    Calculate scaling factors based upon total counts. One of the factors
    is always one (the other is either one or larger than one).

    :arg array_like left, right: A vector.

    :return: A tuple of scaling factors.
    :rtype: float, float
    """
    left_scale = 1.0
    right_scale = 1.0

    # Calculate the scaling factors in such a way that no element is
    # between 0 and 1.
    left_sum = np.sum(left)
    right_sum = np.sum(right)

    if left_sum < right_sum:
        left_scale = right_sum / left_sum
    else:
        right_scale = left_sum / right_sum

    return left_scale, right_scale


def scale_down(left, right):
    """
    Normalise scaling factor between 0 and 1.

    :arg float left, right: Scaling factors.

    :return: Tuple of normalised scaling factors.
    :rtype: float, float
    """
    factor = max(left, right)

    return left / factor, right / factor


def positive(vector, mask):
    """
    Set all zero positions in `mask` to zero in `vector`.

    :arg array_like vector, mask: Vector.

    :return: `vector` with all zero positions in `mask` set to zero.
    :rtype: numpy.ndarray
    """
    return np.multiply(vector, np.asanyarray(mask, dtype=bool))


def multiset(left, right, pairwise):
    """
    Calculate the multiset distance between two vectors.

    :arg array_like left, right: Vector.
    :arg function pairwise: A pairwise distance function.

    :return: The multiset distance between `left` and `right`.
    :rtype: float

    Note that `function` must be vectorized, i.e., it is called directly on
    NumPy arrays, instead of on their pairwise elements. If your function only
    works on individual elements, convert it to a NumPy ufunc first. For
    example::

        >>> f = np.vectorize(f, otypes=['float'])
    """
    left = np.asanyarray(left)
    right = np.asanyarray(right)

    nonzero = np.where(np.logical_or(left, right))
    distances = pairwise(left[nonzero], right[nonzero])
    return distances.sum() / (len(distances) + 1)


def euclidean(left, right):
    """
    Calculate the Euclidean distance between two vectors.

    :arg array_like left, right: Vector.

    :return: The Euclidean distance between `left` and `right`.
    :rtype: float
    """
    return vector_length(np.subtract(left, right))


def cosine_similarity(left, right):
    """
    Calculate the Cosine similarity between two vectors.

    :arg array_like left, right: Vector.

    :return: The Cosine similarity between `left` and `right`.
    :rtype: float
    """
    return np.dot(left, right) / (vector_length(left) * vector_length(right))


#: Vector distance functions.
vector_distance = {
    "default": None,
    "euclidean": euclidean,
    "cosine": cosine_similarity
}


#: Pairwise distance functions. Arguments should be of type `numpy.ndarray`.
pairwise = {
    "prod": lambda x, y: abs(x - y) / ((x + 1) * (y + 1)),
    "sum": lambda x, y: abs(x - y) / (x + y + 1)
}


#: Summary functions.
summary = {
    "min": np.min,
    "average": np.mean,
    "median": np.median
}


#: Merge functions. Arguments should be of type `numpy.ndarray`.
mergers = {
    "sum": lambda x, y: x + y,
    "xor": lambda x, y: (x + y) * np.logical_xor(x, y),
    "int": lambda x, y: x * np.asanyarray(y, dtype=bool),
    "nint": lambda x, y: x * np.logical_not(y)
}



#!/usr/bin/env python

import argparse
import math
import pylab

from scipy import interpolate

def modality(handle, cutoff, resolution):
    """
    """
    x, y = zip(*map(lambda x: map(int, x.split()),
        handle.readlines()[:cutoff]))

    # Ad-hoc weighing function.
    w = map(lambda x: 1.0 / (10 * x + 1), x)
    # Get an approximation of the distribution.
    f = interpolate.UnivariateSpline(x, y, w=w)

    xs = pylab.linspace(0, cutoff, resolution)

    # Plot the original.
    pylab.plot(x, y)

    # Plot the interpolated function.
    ys = f(xs)
    pylab.plot(xs, ys)

    # Plot the tops.
    ys = f(xs, 1)
    g = interpolate.InterpolatedUnivariateSpline(xs, ys)
    pylab.plot(g.roots(), f(g.roots()), 'o')

    # Plot the bending points.
    ys = f(xs, 2)
    g = interpolate.InterpolatedUnivariateSpline(xs, ys)
    pylab.plot(g.roots(), f(g.roots()), 'o')

    pylab.legend(('original', 'interpolated', '1st derivative',
        '2nd derivative'), loc='best')
    pylab.xscale("log")
    pylab.yscale("log")
    pylab.show()
#modality

def main():
    """
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("INPUT", type=argparse.FileType('r'),
        help="distribution file")
    parser.add_argument("-c", dest="cutoff", type=int, default=150,
        help="cutoff value")
    parser.add_argument("-r", dest="resolution", type=int, default=10000,
        help="resolution")

    args = parser.parse_args()

    modality(args.INPUT, args.cutoff, args.resolution)
#main

if __name__ == "__main__":
    main()



"""
*k*-mer profile distance library.

.. moduleauthor:: Leiden University Medical Center <humgen@lumc.nl>
.. moduleauthor:: Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
.. moduleauthor:: Martijn Vermaat <m.vermaat.hg@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import str

import numpy as np

from . import metrics


class ProfileDistance(object):
    """
    Class of distance functions.
    """
    def __init__(self, do_balance=False, do_positive=False, do_smooth=False,
                 summary=metrics.summary['min'], threshold=0, do_scale=False,
                 down=False, distance_function=None,
                 pairwise=metrics.pairwise['prod']):
        """
        Initialise the class.

        :arg bool do_balance: Balance the profiles.
        :arg bool do_positive: Only use positive values.
        :arg bool do_smooth: Use dynamic smoothing.
        :arg function summary: Summary function for dynamic smoothing.
        :arg int threshold: Threshold for the summary function.
        :arg bool do_scale: Scale the profiles.
        :arg bool down: Normalise the scaling factors between 0 and 1.
        :arg function distance_function: Use a specific distance function.
        :arg function pairwise: Pairwise distance function for the multiset
          distance.
        """
        self._do_balance = do_balance
        self._do_positive = do_positive
        self._do_smooth = do_smooth
        self._threshold = threshold
        self._do_scale = do_scale
        self._down = down
        self._distance_function = distance_function
        self._pairwise = pairwise
        self._function = summary

    def _collapse(self, vector, start, length):
        """
        Collapse a part of *k*-mer counts into a list of four numbers,
        representing the *k*-mers that start with a particular letter.

        :arg vector: Counts of a *k*-mer profile.
        :type vector: array_like, 1 dimension
        :arg start: Start of the area to collapse.
        :type start: int
        :arg length: Length of the area to collapse.
        :type length: int

        :return: Collapsed sub-profile.
        :rtype: ndarray
        """
        area = vector[start:start + length]
        return np.reshape(area, (4, length // 4)).sum(axis=1)

    def _dynamic_smooth(self, left, right, start, length):
        """
        Smooth two profiles by collapsing sub-profiles that do not meet the
        requirements governed by the selected summary function and the
        threshold.

        :arg kpal.klib.Profile left, right: Profile to smooth.
        :arg int start: Start of the sub-profile to smooth.
        :arg int length: Length of the sub-profile to smooth.
        """
        # If we use function=min and threshold=0, we should get the following
        # transformation:
        #
        #           | before           | after
        # ----------+------------------+-----------------
        #           | 0111111111111011 | 3000111111113000
        # profile A | ACGTACGTACGTACGT | ACGTACGTACGTACGT
        #           | AAAACCCCGGGGTTTT | AAAACCCCGGGGTTTT
        # ----------+------------------+-----------------
        #           | 0101111111111111 | 2000111111114000
        # profile B | ACGTACGTACGTACGT | ACGTACGTACGTACGT
        #           | AAAACCCCGGGGTTTT | AAAACCCCGGGGTTTT
        if length == 1:
            return

        left_c = self._collapse(left.counts, start, length)
        right_c = self._collapse(right.counts, start, length)

        if min(self._function(left_c),
               self._function(right_c)) <= self._threshold:
            # Collapse the sub-profile.
            left.counts[start] = left_c.sum()
            right.counts[start] = right_c.sum()

            # Remove the k-mer counts used to collapse.
            left.counts[start + 1:start + length] = 0
            right.counts[start + 1:start + length] = 0

            return

        new_length = length // 4
        for i in range(4):
            self._dynamic_smooth(left, right, start + i * new_length,
                                 new_length)

    def dynamic_smooth(self, left, right):
        """
        Smooth two profiles by collapsing sub-profiles that do not meet the
        requirements governed by the selected summary function and the
        threshold.

        :arg kpal.klib.Profile left, right: Profiles to smooth.
        """
        self._dynamic_smooth(left, right, 0, left.number)

    def distance(self, left, right):
        """
        Calculate the distance between two *k*-mer profiles.

        :arg kpal.klib.Profile left, right: Profiles to calculate distance
          between.

        :return: The distance between `left` and `right`.
        :rtype: float
        """
        left = left.copy()
        right = right.copy()

        if self._do_balance:
            left.balance()
            right.balance()

        if self._do_positive:
            left.counts = metrics.positive(left.counts, right.counts)
            right.counts = metrics.positive(right.counts, left.counts)

        if self._do_smooth:
            self.dynamic_smooth(left, right)
        if self._do_scale:
            left_scale, right_scale = metrics.get_scale(left.counts,
                                                        right.counts)

            if self._down:
                left_scale, right_scale = metrics.scale_down(left_scale,
                                                             right_scale)
            left.counts = left.counts * left_scale
            right.counts = right.counts * right_scale

        if not self._distance_function:
            return metrics.multiset(left.counts, right.counts, self._pairwise)
        return self._distance_function(left.counts, right.counts)


def distance_matrix(profiles, output, precision, dist):
    """
    Make a distance matrix for any number of *k*-mer profiles.

    :arg list(Profile) profiles: List of profiles.
    :arg output: Open writable file handle.
    :type output: file-like object
    :arg int precision: Number of digits in the output.
    :arg kpal.kdistlib.ProfileDistance dist: A distance functions object.
    """
    input_count = len(profiles)

    print(str(input_count), file=output)
    for i in profiles:
        print(i.name, file=output)
    for i in range(1, input_count):
        for j in range(i):
            if (j):
                output.write(' ')
            output.write('{{0:.{0}f}}'.format(precision).format(
                         dist.distance(profiles[i], profiles[j])))

        output.write('\n')



"""
*k*-mer base library.

.. moduleauthor:: Leiden University Medical Center <humgen@lumc.nl>
.. moduleauthor:: Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
.. moduleauthor:: Martijn Vermaat <m.vermaat.hg@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import next, range, str

import itertools
import math
import re

from Bio import SeqIO
import numpy as np

from . import metrics


class Profile(object):
    """
    A *k*-mer profile provides *k*-mer counts and operations on them.

    Instead of using the :class:`Profile` constructor directly, you should
    generally use one of the profile construction methods:

    - :meth:`~Profile.from_file`
    - :meth:`~Profile.from_file_old_format`
    - :meth:`~Profile.from_fasta`

    :arg counts: Array of integers where each element is the count for a
      *k*-mer. Ordering is alphabetically by the *k*-mer.
    :type counts: numpy.ndarray
    :arg str name: Profile name.
    """
    #: Conversion table form nucleotide to binary.
    _nucleotide_to_binary = {
        'A': 0x00, 'a': 0x00,
        'C': 0x01, 'c': 0x01,
        'G': 0x02, 'g': 0x02,
        'T': 0x03, 't': 0x03
    }

    #: Conversion table form binary to nucleotide.
    _binary_to_nucleotide = {
        0x00: 'A',
        0x01: 'C',
        0x02: 'G',
        0x03: 'T'
    }

    def __init__(self, counts, name=None):
        self.length = int(math.log(len(counts), 4))
        self.counts = counts
        self.name = name

    @classmethod
    def from_file(cls, handle, name=None):
        """
        Load the *k*-mer profile from a file.

        :arg h5py.File handle: Open readable *k*-mer profile file handle.
        :arg str name: Profile name.

        :return: A *k*-mer profile.
        :rtype: Profile
        """
        name = name or sorted(handle['profiles'].keys())[0]
        counts = handle['profiles/' + name][:]
        return cls(counts, name=name)

    @classmethod
    def from_file_old_format(cls, handle, name=None):
        """
        Load the *k*-mer profile from a file in the old plaintext format.

        :arg handle: Open readable *k*-mer profile file handle (old format).
        :type handle: file-like object
        :arg str name: Profile name.

        :return: A *k*-mer profile.
        :rtype: Profile
        """
        # Ignore lines with length, total, nonzero.
        for _ in range(3):
            next(handle)

        counts = np.loadtxt(handle, dtype='int64')
        return cls(counts, name=name)

    @classmethod
    def from_fasta(cls, handle, length, name=None):
        """
        Create a *k*-mer profile from a FASTA file by counting all *k*-mers in
        each line.

        :arg handle: Open readable FASTA file handle.
        :type handle: file-like object
        :arg int length: Length of the *k*-mers.
        :arg str name: Profile name.

        :return: A *k*-mer profile.
        :rtype: Profile
        """
        sequences = (str(record.seq) for record in SeqIO.parse(handle, 'fasta'))
        return cls.from_sequences(sequences, length, name=name)

    @classmethod
    def from_fasta_by_record(cls, handle, length, prefix=None):
        """
        Create *k*-mer profiles from a FASTA file by counting all *k*-mers per
        record. Profiles are named by the record names.

        :arg handle: Open readable FASTA file handle.
        :type handle: file-like object
        :arg int length: Length of the *k*-mers.
        :arg str prefix: If provided, the names of the *k*-mer profiles are
          prefixed with this.

        :return: A generator yielding the created *k*-mer profiles.
        :rtype: iterator(Profile)
        """
        prefix = prefix + '_' if prefix else ''

        for i, record in enumerate(SeqIO.parse(handle, 'fasta')):
            name = prefix + (record.name or str(i + 1))
            yield cls.from_sequences([str(record.seq)], length, name=name)

    @classmethod
    def from_sequences(cls, sequences, length, name=None):
        """
        Create a *k*-mer profile from `sequences` by counting all *k*-mers in
        each sequence.

        :arg sequences: An iterable of string sequences.
        :type sequences: iterator(str)
        :arg int length: Length of the *k*-mers.
        :arg str name: Profile name.

        :return: A *k*-mer profile.
        :rtype: Profile
        """
        number = 4 ** length
        bitmask = number - 0x01
        counts = [0] * number
        alphabet = re.compile('[^' + ''.join(cls._nucleotide_to_binary) + ']')

        for sequence in sequences:
            for part in alphabet.split(sequence):
                if len(part) >= length:
                    binary = 0x00

                    # Calculate the binary representation of a k-mer.
                    for i in part[:length]:
                        binary = (binary << 2) | cls._nucleotide_to_binary[i]
                    counts[binary] += 1

                    # Calculate the binary representation of the next k-mer.
                    for i in part[length:]:
                        binary = ((binary << 2) |
                                  cls._nucleotide_to_binary[i]) & bitmask
                        counts[binary] += 1

        return cls(np.array(counts, dtype='int64'), name=name)

    @property
    def name(self):
        """
        Profile name.
        """
        return self._name

    @name.setter
    def name(self, name):
        if name and ('/' in name or '.' in name):
            raise ValueError('Profile name may not contain / or . characters.')
        self._name = name

    @property
    def number(self):
        """
        Number of possible *k*-mers with this length.
        """
        return len(self.counts)

    @property
    def non_zero(self):
        """
        Number *k*-mers with a non-zero count.
        """
        return np.count_nonzero(self.counts)

    @property
    def total(self):
        """
        Sum of *k*-mer counts.
        """
        return self.counts.sum()

    @property
    def mean(self):
        """
        Mean of *k*-mer counts.
        """
        return self.counts.mean()

    @property
    def median(self):
        """
        Median of *k*-mer counts.
        """
        return np.median(self.counts)

    @property
    def std(self):
        """
        Standard deviation of *k*-mer counts.
        """
        return self.counts.std()

    def save(self, handle, name=None):
        """
        Save the *k*-mer counts to a file.

        :arg h5py.File handle: Open writeable *k*-mer profile file handle.
        :arg name: Profile name in the file. If not provided, the current
          profile name is used, or the first available number from 1
          consecutively if the profile has no name.
        :type name: str

        :return: Profile name in the file.
        :rtype: str
        """
        if name and ('/' in name or '.' in name):
            raise ValueError('Profile name may not contain / or . characters.')

        name = name or self.name or next(str(n) for n in itertools.count(1)
                                         if str(n) not in handle['profiles'])

        profile = handle.create_dataset('profiles/' + name, data=self.counts,
                                        dtype='int64', compression='gzip')
        profile.attrs['length'] = self.length
        profile.attrs['total'] = self.total
        profile.attrs['non_zero'] = self.non_zero
        profile.attrs['mean'] = self.mean
        profile.attrs['median'] = self.median
        profile.attrs['std'] = self.std
        handle.flush()

        return name

    def copy(self):
        """
        Create a copy of the *k*-mer profile. This returns a deep copy, so
        modifying the copy's *k*-mer counts will not affect the original and
        vice versa.

        :return: Deep copy of profile.
        :rtype: Profile
        """
        return type(self)(self.counts.copy(), name=self.name)

    def merge(self, profile, merger=metrics.mergers["sum"]):
        """
        Merge two profiles.

        :arg Profile profile: Another *k*-mer profile.
        :arg function merger: A pairwise merge function.

        Note that `function` must be vectorized, i.e., it is called directly
        on NumPy arrays, instead of on their pairwise elements. If your
        function only works on individual elements, convert it to a NumPy
        ufunc first. For example::

            >>> f = np.vectorize(f, otypes=['int64'])
        """
        self.counts = merger(self.counts, profile.counts)

    def balance(self):
        """
        Add the counts of the reverse complement of a *k*-mer to the *k*-mer
        and vice versa.
        """
        for i in range(self.number):
            i_rc = self.reverse_complement(i)

            if i < i_rc:
                temp = self.counts[i]
                self.counts[i] += self.counts[i_rc]
                self.counts[i_rc] += temp
            elif i == i_rc:
                self.counts[i] += self.counts[i]

    def split(self):
        """
        Split the profile into two lists, every position in the first list has
        its reverse complement in the same position in the second list and
        vice versa. All counts are doubled, so we can equaly distribute
        palindrome counts over both lists.

        Note that the returned counts are not *k*-mer profiles. They can be
        used to show the balance of the original profile by calculating the
        distance between them.

        :return: The doubled forward and reverse complement counts.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        forward = []
        reverse = []

        for i in range(self.number):
            i_rc = self.reverse_complement(i)

            if i < i_rc:
                forward.append(self.counts[i] * 2)
                reverse.append(self.counts[i_rc] * 2)
            elif i == i_rc:
                forward.append(self.counts[i])
                reverse.append(self.counts[i])

        return np.array(forward), np.array(reverse)

    def shrink(self, factor=1):
        """
        Shrink the profile, effectively reducing the value of *k*.

        Note that this operation may give slightly different values than
        counting at a lower *k* directly.

        :arg int factor: Shrinking factor.
        """
        if self.length <= factor:
            raise ValueError(
                "Reduction factor should be smaller than k-mer size.")

        # I don't know how to do this operation directly with NumPy
        # instructions without creating a complete copy of the counts in
        # memory (e.g., using `np.split(self.counts, ...)`).
        # So instead we just build the new counts one at a time. Note that
        # the builtin Python `sum` is actually faster than `np.sum` here,
        # since the arrays are very small.
        merge_size = 4 ** factor
        new_counts = (sum(self.counts[i:i + merge_size])
                      for i in range(0, self.number, merge_size))
        self.counts = np.fromiter(new_counts, dtype='int64')
        self.length -= factor

    def shuffle(self):
        """
        Randomise the profile.
        """
        np.random.shuffle(self.counts)

    def dna_to_binary(self, sequence):
        """
        Convert a string of DNA to an integer.

        :arg str sequence: DNA sequence.

        :return: Binary representation of `sequence`.
        :rtype: int
        """
        result = 0x00

        for i in sequence:
            result <<= 2
            result |= self._nucleotide_to_binary[i]

        return result

    def binary_to_dna(self, number):
        """
        Convert an integer to a DNA string.

        :arg int number: Binary representation of a DNA sequence.

        :returns: DNA string corresponding to `number`.
        :rtype: str
        """
        sequence = ""

        for i in range(self.length):
            sequence += self._binary_to_nucleotide[number & 0x03]
            number >>= 2

        return sequence[::-1]

    def reverse_complement(self, number):
        """
        Calculate the reverse complement of a DNA sequence in a binary
        representation.

        :arg int number: Binary representation of a DNA sequence.

        :return: Binary representation of the reverse complement of the
          sequence corresponding to `number`.
        :rtype: int
        """
        number = ~number
        result = 0x00

        for i in range(self.length):
            result = (result << 2) | (number & 0x03)
            number >>= 2

        return result

    def _ratios_matrix(self):
        """
        Calculate all relative frequencies of *k*-mers. If a division by 0
        occurs, the frequency will be set to -1.0.

        :return: A matrix with relative frequencies.
        :rtype: float[][]
        """
        # Perhaps we can use something like this for a future normalization
        # operation.
        # Todo: Do this directly with NumPy.
        ratios = []
        for i in range(self.number):
            ratios.append(self.number * [0.0])

        # Fill the matrix.
        for i in range(self.number):
            for j in range(self.number):
                if self.counts[j]:
                    ratios[i][j] = (self.counts[i] /
                                    self.counts[j]) / self.total
                else:
                    ratios[i][j] = -1.0

        return ratios

    def _freq_diff_matrix(self):
        """
        Calculate all frequency differences of *k*-mers.

        :return: A matrix with frequency differences.
        :rtype: float[][]
        """
        ratios = []
        for i in range(self.number):
            ratios.append(self.number * [0])

        # Fill the matrix.
        for i in range(self.number):
            for j in range(self.number):
                if self.counts[j]:
                    ratios[i][j] = abs(self.counts[i] -
                                       self.counts[j]) / self.total

        return ratios

    def print_counts(self):
        """
        Print the *k*-mer counts.
        """
        for i in range(self.number):
            print(self.binary_to_dna(i), self.counts[i])

    def _print_ratios(self, ratios):
        """
        Print a ratios matrix.

        :arg ratios: A matrix with relative frequencies.
        :type ratios: float[][]
        """
        # The header.
        print((self.length + 1) * ' ', end=' ')
        for i in range(self.number):
            print(self.binary_to_dna(i), end='   ')
        print()

        # The matrix.
        for i in range(self.number):
            print(self.binary_to_dna(i), end=' ')
            for j in range(self.number):

                print('{{0:.{0}f}}'.format(self.length).format(ratios[i][j]),
                      end=' ')
            print()



import os
from setuptools import setup
import sys

if sys.version_info < (2, 6):
    raise Exception('kPAL requires Python 2.6 or higher.')

dependencies = ['numpy', 'h5py>=2.1', 'biopython', 'semantic-version',
                'future']
if sys.version_info[:2] == (2, 6):
    dependencies.extend(['argparse', 'importlib', 'unittest2'])

# This is quite the hack, but we don't want to import our package from here
# since that's recipe for disaster (it might have some uninstalled
# dependencies, or we might import another already installed version).
distmeta = {}
for line in open(os.path.join('kpal', '__init__.py')):
    try:
        field, value = (x.strip() for x in line.split('='))
    except ValueError:
        continue
    if field == '__version_info__':
        value = value.strip('[]()')
        value = '.'.join(x.strip(' \'"') for x in value.split(','))
    else:
        value = value.strip('\'"')
    distmeta[field] = value

try:
    with open('README.rst') as readme:
        long_description = readme.read()
except IOError:
    long_description = 'See ' + distmeta['__homepage__']

setup(
    name='kPAL',
    version=distmeta['__version_info__'],
    description='k-mer analysis toolkit and programming library.',
    long_description=long_description,
    author=distmeta['__author__'],
    author_email=distmeta['__contact__'],
    url=distmeta['__homepage__'],
    license='MIT License',
    platforms=['any'],
    packages=['kpal'],
    install_requires=dependencies,
    entry_points={
        'console_scripts': ['kpal = kpal.kmer:main']
        },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
    keywords='bioinformatics'
)



"""
kPAL (*k*-mer profile analysis library)


Analysis toolkit and programming library for k-mer profiles.

Copyright (c) 2013-2014 Leiden University Medical Center <humgen@lumc.nl>
Copyright (c) 2013-2014 Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
Copyright (c) 2014 Martijn Vermaat <m.vermaat.hg@lumc.nl>

Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import str
from future.utils import native

import argparse
from io import open
import os
import sys

import h5py
import semantic_version


__version_info__ = ('2', '1', '1')
__date__ = '14 Aug 2015'


__version__ = '.'.join(__version_info__)
__author__ = 'LUMC, Jeroen F.J. Laros, Martijn Vermaat'
__contact__ = 'J.F.J.Laros@lumc.nl'
__homepage__ = 'https://git.lumc.nl/j.f.j.laros/k-mer'


USAGE = __doc__.split("\n\n\n")
FORMAT_VERSION = semantic_version.Version('1.0.0')
FORMAT_ACCEPT = semantic_version.Spec('>=1.0.0,<2.0.0')


# Same as argparse.FileType in Python 3, but using io.open to get the same
# behaviour on Python 2, and adding protection against overwriting existing
# files.
class FileType(object):
    def __init__(self, mode='r', bufsize=-1, encoding=None, errors=None):
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = encoding
        self._errors = errors

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                return sys.stdin
            elif 'w' in self._mode:
                return sys.stdout
            else:
                msg = 'argument "-" with mode %r' % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            if 'w' in self._mode and os.path.exists(string):
                raise OSError('file exists')
            return open(string, self._mode, self._bufsize, self._encoding,
                        self._errors)
        except OSError as e:
            message = "can't open '%s': %s"
            raise argparse.ArgumentTypeError(message % (string, e))

    def __repr__(self):
        args = self._mode, self._bufsize
        kwargs = [('encoding', self._encoding), ('errors', self._errors)]
        args_str = ', '.join([repr(arg) for arg in args if arg != -1] +
                             ['%s=%r' % (kw, arg) for kw, arg in kwargs
                              if arg is not None])
        return '%s(%s)' % (type(self).__name__, args_str)


class ProfileFileType(object):
    def __init__(self, mode='r'):
        self._mode = mode

    def __call__(self, string):
        try:
            if 'w' in self._mode and os.path.exists(string):
                raise IOError('file exists')
            handle = h5py.File(string, self._mode)
            if 'w' in self._mode:
                handle.attrs['format'] = 'kMer'
                handle.attrs['version'] = native(str(FORMAT_VERSION))
                handle.attrs['producer'] = 'kPAL %s' % __version__
                handle.create_group('profiles')
            else:
                if handle.attrs.get('format') != 'kMer':
                    raise IOError('not a k-mer profile file')
                version = semantic_version.Version(handle.attrs['version'])
                if version not in FORMAT_ACCEPT:
                    raise IOError('file format version %s not supported'
                                  % version)
            return handle
        except IOError as e:
            message = "can't open '%s': %s"
            raise argparse.ArgumentTypeError(message % (string, e))

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self._mode)


def doc_split(func):
    return func.__doc__.split("\n\n")[0]


def version(name):
    return '{0} version {1}\n\nAuthor   : {2} <{3}>\nHomepage : {4}'.format(
        name, __version__, __author__, __contact__, __homepage__)



"""
Toolbox for k-mer profiles.

.. moduleauthor:: Leiden University Medical Center <humgen@lumc.nl>
.. moduleauthor:: Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
.. moduleauthor:: Martijn Vermaat <m.vermaat.hg@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import str, zip

import argparse
import importlib
import os
import re
import sys

import numpy as np

from . import (USAGE, FileType, ProfileFileType, doc_split, version, klib,
               kdistlib, metrics)


LENGTH_ERROR = 'k-mer lengths of the files differ'
NAMES_COUNT_ERROR = ('number of profile names does not match number of '
                     'profiles')
PAIRED_NAMES_COUNT_ERROR = ('number of left and right profile names do not '
                            'match')
PREFIX_COUNT_ERROR = ('number of name prefixes does not match number of '
                      'profiles')


# Importable definition, e.g. `package.module.merge_function`.
# http://docs.python.org/2/reference/lexical_analysis.html#identifiers
_PYTHON_IMPORTABLE = '{0}(\.{0})+$'.format('[_a-zA-Z][_a-zA-Z0-9]*')


def _name_from_handle(handle):
    """
    Try to get a name for `handle` from its filename, if there is one. Return
    `None` for non-filename handles such as string buffers or streams.
    """
    if not hasattr(handle, 'name') or handle.name.startswith('<'):
        return None
    return os.path.splitext(os.path.basename(str(handle.name)))[0]


def convert(input_handles, output_handle, names=None):
    """
    Save k-mer profiles from files in the old plaintext format (used by kPAL
    versions < 1.0.0) to a k-mer profile file in the current HDF5 format.

    :arg input_handles: Open readable k-mer profile file handles (old format).
    :type input_handles: list(file-like object)
    :arg output_handle: Open writeable k-mer profile file handle.
    :type output_handle: h5py.File
    :arg names: Optional list of names for the saved k-mer profiles (must have
      the same length as `input_handles`). If not provided, profiles are named
      according to the input filenames, or numbered consecutively from 1 if no
      filenames are available.
    :type names: list(str)
    """
    names = names or [_name_from_handle(h) for h in input_handles]

    if len(names) != len(input_handles):
        raise ValueError(NAMES_COUNT_ERROR)

    for input_handle, name in zip(input_handles, names):
        profile = klib.Profile.from_file_old_format(input_handle, name=name)
        profile.save(output_handle)


def cat(input_handles, output_handle, names=None, prefixes=None):
    """
    Save k-mer profiles from several files to one k-mer profile file.

    :arg input_handles: Open readable k-mer profile file handles.
    :type input_handles: list(file-like object)
    :arg output_handle: Open writeable k-mer profile file handle.
    :type output_handle: h5py.File
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in all input files are considered.
    :type names: list(str)
    :arg prefixes: Optional list of prefixes to use for the saved k-mer
      profile names, one per input file. If not provided, profile names are
      assumed to be disjoint and no prefix is used.
    :type prefixes: list(str)
    """
    prefixes = prefixes or ['' for _ in input_handles]

    if len(prefixes) != len(input_handles):
        raise ValueError(PREFIX_COUNT_ERROR)

    for input_handle, prefix in zip(input_handles, prefixes):
        names_ = names or sorted(input_handle['profiles'])

        for name in names_:
            try:
                profile = klib.Profile.from_file(input_handle, name=name)
            except KeyError:
                # In this specific case, we ignore non-existing profiles,
                # since the user may have specified them for selecting from
                # one of the other input files.
                continue
            profile.save(output_handle, name=prefix + name)


def count(input_handles, output_handle, size, names=None, by_record=False):
    """
    Make k-mer profiles from FASTA files.

    :arg input_handles: Open readable FASTA file handles.
    :type input_handles: list(file-like object)
    :arg output_handle: Open writeable k-mer profile file handle.
    :type output_handle: h5py.File
    :arg size: Size of k.
    :type size: int
    :arg names: Optional list of names for the created k-mer profiles (must
      have the same length as `input_handles`). If not provided, profiles are
      named according to the input filenames, or numbered consecutively from 1
      if no filenames are available.
    :type names: list(str)
    :arg bool by_record: If `True`, make a k-mer profile per FASTA record
      instead of a k-mer profile per FASTA file. Profiles are named by the
      record names and prefixed according to `names` if more than one FASTA
      file is given).
    """
    names = names or [_name_from_handle(h) for h in input_handles]

    if len(names) != len(input_handles):
        raise ValueError(NAMES_COUNT_ERROR)

    for input_handle, name in zip(input_handles, names):
        if by_record:
            prefix = name if len(input_handles) > 1 else None
            profiles = klib.Profile.from_fasta_by_record(input_handle, size,
                                                         prefix=prefix)
        else:
            profiles = [klib.Profile.from_fasta(input_handle, size, name=name)]

        for profile in profiles:
            profile.save(output_handle)


def merge(input_handle_left, input_handle_right, output_handle,
          names_left=None, names_right=None, merger='sum', custom_merger=None):
    """
    Merge k-mer profiles. If the files contain more than one profile, they are
    linked by name and merged pairwise. The resulting profile name is set to
    that of the original profiles if they match, or to their concatenation
    otherwise.

    :arg h5py.File input_handle_left, input_handle_right: Open readable k-mer
      profile file handle.
    :arg h5py.File output_handle: Open writeable k-mer profile file handle.
    :arg list(str) names_left, names_right: Optional list of names of the
      k-mer profiles to consider. If not provided, all profiles in the file
      are considered.
    :arg function merger: Merge function.
    :arg str custom_merger: Custom merge function.
    """
    names_left = names_left or sorted(input_handle_left['profiles'])
    names_right = names_right or sorted(input_handle_right['profiles'])

    if len(names_left) != len(names_right):
        raise ValueError(PAIRED_NAMES_COUNT_ERROR)

    if custom_merger:
        if re.match(_PYTHON_IMPORTABLE, custom_merger):
            # Importable definition, e.g. `package.module.merge_function`.
            module, name = custom_merger.rsplit('.', 1)
            merge_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `left` and `right`, e.g. `np.add(left, right)`.
            # The `numpy` package is available as `np`.
            merge_function = eval('lambda left, right: ' + custom_merger,
                                  {'np': np})
    else:
        merge_function = metrics.mergers[merger]

    for name_left, name_right in zip(names_left, names_right):
        profile_left = klib.Profile.from_file(input_handle_left,
                                              name=name_left)
        profile_right = klib.Profile.from_file(input_handle_right,
                                               name=name_right)

        if profile_left.length != profile_right.length:
            raise ValueError(LENGTH_ERROR)

        profile_right.merge(profile_left, merge_function)

        if name_left == name_right:
            name = name_left
        else:
            name = name_left + '_' + name_right
        profile_right.save(output_handle, name=name)


def balance(input_handle, output_handle, names=None):
    """
    Balance k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg h5py.File output_handle: Open writeable k-mer profile file handle.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)
        profile.balance()
        profile.save(output_handle)


def get_balance(input_handle, output_handle, precision=10, names=None):
    """
    Show the balance of k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg output_handle: Open writeable k-mer profile file handle.
    :type output_handle: file-like object
    :arg int precision: Number of digits in the output.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    # Todo: Wouldn't it make more sense conceptually to create the reverse
    # complement profile and calculate the distance to that? Perhaps it's
    # even easier to vectorize in NumPy, so faster?
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)

        forward, reverse = profile.split()
        balance = metrics.multiset(forward, reverse,
                                   metrics.pairwise['prod'])
        print(name, '{{0:.{0}f}}'.format(precision).format(balance),
              file=output_handle)


def get_stats(input_handle, output_handle, precision=10, names=None):
    """
    Show the mean and standard deviation of k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg int precision: Number of digits in the output.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)

        print(name,
              '{{0:.{0}f}}'.format(precision).format(profile.mean),
              '{{0:.{0}f}}'.format(precision).format(profile.std),
              file=output_handle)


def distribution(input_handle, output_handle, names=None):
    """
    Calculate the distribution of the values in k-mer profiles. Every output
    line has the name of the profile, the count and the number of k-mers with
    this count.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)

        print('\n'.join('{0} {1} {2}'.format(name, v, c)
                        for v, c in metrics.distribution(profile.counts)),
              file=output_handle)


def info(input_handle, output_handle, names=None):
    """
    Print some information about k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    print('File format version:', input_handle.attrs['version'],
          file=output_handle)
    print('Produced by:', input_handle.attrs['producer'], file=output_handle)

    for name in names:
        # Todo: This loads the entire profile, which we actually do not need.
        #   Especially for files with many profiles, we might want to just
        #   get the statistics without loading the profile.
        profile = klib.Profile.from_file(input_handle, name=name)

        print('', file=output_handle)
        print('Profile:', profile.name, file=output_handle)
        print('- k-mer length:', str(profile.length),
              '({0} k-mers)'.format(profile.number), file=output_handle)
        print('- Zero counts:', str(profile.number - profile.non_zero),
              file=output_handle)
        print('- Non-zero counts:', str(profile.non_zero), file=output_handle)
        print('- Sum of counts:', str(profile.total), file=output_handle)
        print('- Mean of counts:', '{0:.3f}'.format(profile.mean),
              file=output_handle)
        print('- Median of counts:', '{0:.3f}'.format(profile.median),
              file=output_handle)
        print('- Standard deviation of counts:',
              '{0:.3f}'.format(profile.std), file=output_handle)


def get_count(input_handle, output_handle, word, names=None):
    """
    Retrieve the counts in k-mer profiles for a particular word.

    :arg h5py.File input_handle: Open readable handle to a k-mer profile file.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg str word: The query word.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)
        if profile.length != len(word):
            raise ValueError('the length of the query does not match the '
                             'profile length')
        try:
            offset = profile.dna_to_binary(word)
        except KeyError:
            raise ValueError('the input is not a valid DNA sequence')
        print(name, str(profile.counts[offset]), file=output_handle)


def positive(input_handle_left, input_handle_right, output_handle_left,
             output_handle_right, names_left=None, names_right=None):
    """
    Only keep counts that are positive in both k-mer profiles. If the files
    contain more than one profile, they are linked by name and processed
    pairwise.

    :arg h5py.File input_handle_left, input_handle_right: Open readable k-mer
      profile file handle.
    :arg h5py.File output_handle_left, output_handle_right: Open writeable
      k-mer profile file handle.
    :arg list(str) names_left, names_right: Optional list of names of the
      k-mer profiles consider. If not provided, all profiles in the file are
      considered.
    """
    names_left = names_left or sorted(input_handle_left['profiles'])
    names_right = names_right or sorted(input_handle_right['profiles'])

    if len(names_left) != len(names_right):
        raise ValueError(PAIRED_NAMES_COUNT_ERROR)

    for name_left, name_right in zip(names_left, names_right):
        profile_left = klib.Profile.from_file(input_handle_left,
                                              name=name_left)
        profile_right = klib.Profile.from_file(input_handle_right,
                                               name=name_right)

        if profile_left.length != profile_right.length:
            raise ValueError(LENGTH_ERROR)

        profile_left.counts = metrics.positive(profile_left.counts,
                                               profile_right.counts)
        profile_right.counts = metrics.positive(profile_right.counts,
                                                profile_left.counts)

        profile_left.save(output_handle_left)
        profile_right.save(output_handle_right)


def scale(input_handle_left, input_handle_right, output_handle_left,
          output_handle_right, names_left=None, names_right=None, down=False):
    """
    Scale two profiles such that the total number of k-mers is equal. If the
    files contain more than one profile, they are linked by name and processed
    pairwise.

    :arg h5py.File input_handle_left, input_handle_right: Open readable k-mer
      profile handle.
    :arg h5py.File output_handle_left, output_handle_right: Open writeable
      k-mer profile handle.
    :arg list(str) names_left, names_right: Optional list of names of the
      k-mer profiles to consider. If not provided, all profiles in the file
      are considered.
    :arg bool down: Scale down.
    """
    names_left = names_left or sorted(input_handle_left['profiles'])
    names_right = names_right or sorted(input_handle_right['profiles'])

    if len(names_left) != len(names_right):
        raise ValueError(PAIRED_NAMES_COUNT_ERROR)

    for name_left, name_right in zip(names_left, names_right):
        profile_left = klib.Profile.from_file(input_handle_left,
                                              name=name_left)
        profile_right = klib.Profile.from_file(input_handle_right,
                                               name=name_right)

        if profile_left.length != profile_right.length:
            raise ValueError(LENGTH_ERROR)

        scale_left, scale_right = metrics.get_scale(profile_left.counts,
                                                    profile_right.counts)
        if down:
            scale_left, scale_right = metrics.scale_down(scale_left,
                                                         scale_right)
        profile_left.counts = profile_left.counts * scale_left
        profile_right.counts = profile_right.counts * scale_right

        profile_left.save(output_handle_left)
        profile_right.save(output_handle_right)


def shrink(input_handle, output_handle, factor, names=None):
    """
    Shrink k-mer profiles, effectively reducing k.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg h5py.File output_handle: Open writeable k-mer profile file handle.
    :arg int factor: Scaling factor.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)
        profile.shrink(factor)
        profile.save(output_handle)


def shuffle(input_handle, output_handle, names=None):
    """
    Randomise k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg h5py.File output_handle: Open writeable k-mer profile file handle.
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    """
    names = names or sorted(input_handle['profiles'])

    for name in names:
        profile = klib.Profile.from_file(input_handle, name=name)
        profile.shuffle()
        profile.save(output_handle)


def smooth(input_handle_left, input_handle_right, output_handle_left,
           output_handle_right, names_left=None, names_right=None,
           summary='min', custom_summary=None, threshold=0):
    """
    Smooth two profiles by collapsing sub-profiles. If the files contain more
    than one profile, they are linked by name and processed pairwise.

    :arg h5py.File input_handle_left, input_handle_right: Open readable k-mer
      profile handle.
    :arg h5py.File output_handle_left, output_handle_right: Open writeable
      k-mer profile file handle.
    :arg list(str) names_left, names_right: Optional list of names of the
      k-mer profiles to consider. If not provided, all profiles in the file
      are considered.
    :arg str summary: Name of the summary function.
    :arg str custom_summary: Custom summary function.
    :arg int threshold: Threshold for the summary function.
    """
    names_left = names_left or sorted(input_handle_left['profiles'])
    names_right = names_right or sorted(input_handle_right['profiles'])

    if len(names_left) != len(names_right):
        raise ValueError(PAIRED_NAMES_COUNT_ERROR)

    if custom_summary:
        if re.match(_PYTHON_IMPORTABLE, custom_summary):
            # Importable definition, e.g. `package.module.summary_function`.
            module, name = custom_summary.rsplit('.', 1)
            summary_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `values`, e.g. `np.max(values)`. The `numpy`
            # package is available as `np`.
            summary_function = eval('lambda values: ' + custom_summary,
                                    {'np': np})
    else:
        summary_function = metrics.summary[summary]

    dist = kdistlib.ProfileDistance(summary=summary_function,
                                    threshold=threshold)

    for name_left, name_right in zip(names_left, names_right):
        profile_left = klib.Profile.from_file(input_handle_left,
                                              name=name_left)
        profile_right = klib.Profile.from_file(input_handle_right,
                                               name=name_right)

        if profile_left.length != profile_right.length:
            raise ValueError(LENGTH_ERROR)

        dist.dynamic_smooth(profile_left, profile_right)

        profile_left.save(output_handle_left)
        profile_right.save(output_handle_right)


def distance(input_handle_left, input_handle_right, output_handle,
             names_left=None, names_right=None, distance_function='default',
             pairwise='prod', custom_pairwise=None, do_smooth=False,
             summary='min', custom_summary=None, threshold=0, do_scale=False,
             down=False, do_positive=False, do_balance=False, precision=10):
    """
    Calculate the distance between two k-mer profiles. If the files contain
    more than one profile, they are linked by name and processed pairwise.

    :arg h5py.File input_handle_left, input_handle_right: Open readable k-mer
      profile handle.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg list(str) names_left, names_right: Optional list of names of the
      k-mer profiles to consider. If not provided, all profiles in the file
      are considered.
    :arg str distance_function: Use a specific distance function.
    :arg str pairwise: Name of the pairwise distance function.
    :arg str custom_pairwise: Custom pairwise distance function.
    :arg bool do_smooth: Enable smoothing.
    :arg str summary: Name of the summary function.
    :arg str custom_summary: Custom summary function.
    :arg int threshold: Threshold for the summary function.
    :arg bool do_scale: Scale the profiles.
    :arg bool down: Scale down.
    :arg bool do_positive: Only use positive values.
    :arg bool do_balance: Balance the profiles.
    :arg int precision: Number of digits in the output.
    """
    names_left = names_left or sorted(input_handle_left['profiles'])
    names_right = names_right or sorted(input_handle_right['profiles'])

    if len(names_left) != len(names_right):
        raise ValueError(PAIRED_NAMES_COUNT_ERROR)

    if custom_summary:
        if re.match(_PYTHON_IMPORTABLE, custom_summary):
            # Importable definition, e.g. `package.module.summary_function`.
            module, name = custom_summary.rsplit('.', 1)
            summary_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `values`, e.g. `np.max(values)`. The `numpy`
            # package is available as `np`.
            summary_function = eval('lambda values: ' + custom_summary,
                                    {'np': np})
    else:
        summary_function = metrics.summary[summary]

    if custom_pairwise:
        if re.match(_PYTHON_IMPORTABLE, custom_pairwise):
            # Importable definition, e.g. `package.module.pairwise_function`.
            module, name = custom_pairwise.rsplit('.', 1)
            pairwise_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `left` and `right`, e.g. `np.add(left, right)`.
            # The `numpy` package is available as `np`.
            pairwise_function = eval('lambda left, right: ' + custom_pairwise,
                                     {'np': np})
    else:
        pairwise_function = metrics.pairwise[pairwise]

    dist = kdistlib.ProfileDistance(
        do_balance=do_balance, do_positive=do_positive, do_smooth=do_smooth,
        summary=summary_function, threshold=threshold, do_scale=do_scale,
        down=down, pairwise=pairwise_function,
        distance_function=metrics.vector_distance[distance_function])

    for name_left, name_right in zip(names_left, names_right):
        profile_left = klib.Profile.from_file(input_handle_left,
                                              name=name_left)
        profile_right = klib.Profile.from_file(input_handle_right,
                                               name=name_right)

        if profile_left.length != profile_right.length:
            raise ValueError(LENGTH_ERROR)

        print(name_left, name_right,
              '{{0:.{0}f}}'.format(precision).format(dist.distance(
                  profile_left, profile_right)),
              file=output_handle)


def distance_matrix(input_handle, output_handle, names=None,
                    distance_function='default', pairwise='prod',
                    custom_pairwise=None, do_smooth=False, summary='min',
                    custom_summary=None, threshold=0, do_scale=False,
                    down=False, do_positive=False, do_balance=False,
                    precision=10):
    """
    Make a distance matrix between any number of k-mer profiles.

    :arg h5py.File input_handle: Open readable k-mer profile file handle.
    :arg output_handle: Open writeable file handle.
    :type output_handle: file-like object
    :arg names: Optional list of names of the k-mer profiles to consider. If
      not provided, all profiles in the file are considered in alphabetical
      order.
    :type names: list(str)
    :arg bool euclidean: Use the Euclidean distance.
    :arg str pairwise: Name of the pairwise distance function.
    :arg str custom_pairwise: Custom pairwise distance function.
    :arg bool do_smooth: Enable smoothing.
    :arg str summary: Name of the summary function.
    :arg str custom_summary: Custom summary function.
    :arg int threshold: Threshold for the summary function.
    :arg bool do_scale: Scale the profiles.
    :arg bool down: Scale down.
    :arg bool do_positive: Only use positive values.
    :arg bool do_balance: Balance the profiles.
    :arg int precision: Number of digits in the output.
    """
    names = names or sorted(input_handle['profiles'])

    if len(names) < 2:
        raise ValueError('you must give at least two k-mer profiles')

    if custom_summary:
        if re.match(_PYTHON_IMPORTABLE, custom_summary):
            # Importable definition, e.g. `package.module.summary_function`.
            module, name = custom_summary.rsplit('.', 1)
            summary_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `values`, e.g. `np.max(values)`. The `numpy`
            # package is available as `np`.
            summary_function = eval('lambda values: ' + custom_summary,
                                    {'np': np})
    else:
        summary_function = metrics.summary[summary]

    if custom_pairwise:
        if re.match(_PYTHON_IMPORTABLE, custom_pairwise):
            # Importable definition, e.g. `package.module.pairwise_function`.
            module, name = custom_pairwise.rsplit('.', 1)
            pairwise_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `left` and `right`, e.g. `np.add(left, right)`.
            # The `numpy` package is available as `np`.
            pairwise_function = eval('lambda left, right: ' + custom_pairwise,
                                     {'np': np})
    else:
        pairwise_function = metrics.pairwise[pairwise]

    dist = kdistlib.ProfileDistance(
        do_balance=do_balance, do_positive=do_positive, do_smooth=do_smooth,
        summary=summary_function, threshold=threshold, do_scale=do_scale,
        down=down, pairwise=pairwise_function,
        distance_function=metrics.vector_distance[distance_function])

    # Todo: This first creates all profiles in memory, which may be
    #   problematic for large k. We could provide an option to keep only the
    #   two current profiles in memory. We may combine this with an option to
    #   paralellize this function. The downside is that profiles have to be
    #   read from file on each use.
    counts = []
    for name in names:
        counts.append(klib.Profile.from_file(input_handle, name=name))
        if counts[0].length != counts[-1].length:
            raise ValueError(LENGTH_ERROR)

    kdistlib.distance_matrix(counts, output_handle, precision, dist)


def main(args=None):
    """
    Command line interface.

    :arg args: Arguments passed to :meth:`argparse.ArgumentParser.parse_args`
      (default is to use `sys.argv[1:]`).
    :type args: list(str)
    """
    multi_input_parser = argparse.ArgumentParser(add_help=False)
    multi_input_parser.add_argument(
        'input_handles', metavar='INPUT', type=FileType('r'), nargs='*',
        default=[sys.stdin], help='input file (default: stdin)')

    input_profile_parser = argparse.ArgumentParser(add_help=False)
    input_profile_parser.add_argument(
        'input_handle', metavar='INPUT', type=ProfileFileType('r'),
        help='input k-mer profile file',)
    input_profile_parser.add_argument(
        '-p', '--profiles', dest='names', metavar='NAME', type=str, nargs='+',
        help='names of the k-mer profiles to consider (default: all profiles '
        'in INPUT, in alphabetical order)')

    multi_input_profile_parser = argparse.ArgumentParser(add_help=False)
    multi_input_profile_parser.add_argument(
        'input_handles', metavar='INPUT', type=ProfileFileType('r'),
        nargs='+', help='input k-mer profile file')
    multi_input_profile_parser.add_argument(
        '-p', '--profiles', dest='names', metavar='NAME', type=str, nargs='+',
        help='names of the k-mer profiles to consider (default: all profiles '
        'per INPUT, in alphabetical order)')

    paired_input_profile_parser = argparse.ArgumentParser(add_help=False)
    paired_input_profile_parser.add_argument(
        'input_handle_left', metavar='INPUT_LEFT', type=ProfileFileType('r'),
        help='input k-mer profile file (left)')
    paired_input_profile_parser.add_argument(
        'input_handle_right', metavar='INPUT_RIGHT',
        type=ProfileFileType('r'), help='input k-mer profile file (right)')
    paired_input_profile_parser.add_argument(
        '-l', '--profiles-left', dest='names_left', metavar='NAME', type=str,
        nargs='+', help='names of the k-mer profiles to consider (left) '
        '(default: all profiles in INPUT_LEFT, in alphabetical order')
    paired_input_profile_parser.add_argument(
        '-r', '--profiles-right', dest='names_right', metavar='NAME',
        type=str, nargs='+', help='names of the k-mer profiles to consider '
        '(right) (default: all profiles in INPUT_RIGHT, in alphabetical '
        'order)')

    output_parser = argparse.ArgumentParser(add_help=False)
    output_parser.add_argument(
        'output_handle', metavar='OUTPUT', type=FileType('w'),
        help='output file')

    output_profile_parser = argparse.ArgumentParser(add_help=False)
    output_profile_parser.add_argument(
        'output_handle', metavar='OUTPUT', type=ProfileFileType('w'),
        help='output k-mer profile file')

    paired_output_profile_parser = argparse.ArgumentParser(add_help=False)
    paired_output_profile_parser.add_argument(
        'output_handle_left', metavar='OUTPUT_LEFT',
        type=ProfileFileType('w'), help='output k-mer profile file (left)')
    paired_output_profile_parser.add_argument(
        'output_handle_right', metavar='OUTPUT_RIGHT',
        type=ProfileFileType('w'), help='output k-mer profile file (right)')

    scale_parser = argparse.ArgumentParser(add_help=False)
    scale_parser.add_argument(
        '-d', dest='down', action='store_true', help='scale down')

    smooth_parser = argparse.ArgumentParser(add_help=False)
    smooth_parser.add_argument(
        '-s', dest='summary', type=str, default='min',
        choices=metrics.summary, help='summary function for dynamic '
        'smoothing (default: %(default)s)')
    smooth_parser.add_argument(
        '-M', '--custom-summary', metavar='STRING', type=str,
        dest='custom_summary',
        help='custom Python summary function, specified either by an '
        'expression over the NumPy ndarray "values" (e.g., '
        '"np.max(values)"), or an importable name (e.g., '
        '"package.module.summary") that can be called with an ndarray as '
        'argument')
    smooth_parser.add_argument(
        '-t', dest='threshold', metavar='INT', type=int, default=0,
        help='threshold for the summary function (default: %(default)s)')

    precision_parser = argparse.ArgumentParser(add_help=False)
    precision_parser.add_argument(
        '-n', metavar='INT', dest='precision', type=int, default=10,
        help='precision in number of decimals (default: %(default)s)')

    dist_parser = argparse.ArgumentParser(
        add_help=False,
        parents=[scale_parser, smooth_parser, precision_parser])
    dist_parser.add_argument(
        '-b', '--balance', dest='do_balance', action='store_true',
        help='balance the profiles')
    dist_parser.add_argument(
        '--positive', dest='do_positive', action='store_true',
        help='use only positive values')
    dist_parser.add_argument(
        '-S', '--scale', dest='do_scale', action='store_true',
        help='scale the profiles')
    dist_parser.add_argument(
        '-m', '--smooth', dest='do_smooth', action='store_true',
        help='smooth the profiles')
    dist_parser.add_argument(
        '-D', dest='distance_function', type=str, default='default',
        choices=metrics.vector_distance,
        help='choose distance function (default: %(default)s)')
    dist_parser.add_argument(
        '-P', dest='pairwise', type=str, default='prod',
        choices=metrics.pairwise, help='paiwise distance function for the '
        'multiset distance (default: %(default)s)')
    # Todo: Note in the documentation that the custom pairwise must be
    #   vectorized with a tip for using np.vectorize if it is not.
    dist_parser.add_argument(
        '-f', '--pairwise-function', metavar='STRING', dest='custom_pairwise',
        type=str,
        help='custom Python pairwise function, specified either by an '
        'expression over the two NumPy ndarrays "left" and "right" (e.g., '
        '"abs(left - right) / (left + right + 1)"), or an importable name '
        '(e.g., "package.module.pairwise") that can be called with two '
        'ndarrays as arguments')

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=USAGE[0], epilog=USAGE[1])
    parser.add_argument('-v', action='version', version=version(parser.prog))
    subparsers = parser.add_subparsers()

    parser_convert = subparsers.add_parser(
        'convert', parents=[multi_input_parser, output_profile_parser],
        description=doc_split(convert))
    parser_convert.add_argument(
        '-p', '--profiles', dest='names', metavar='NAME', type=str, nargs='+',
        help='names for the saved k-mer profiles, one per INPUT (default: '
        'profiles are named according to the input filenames, or numbered '
        'consecutively from 1 if no filenames are available)')
    parser_convert.set_defaults(func=convert)

    parser_cat = subparsers.add_parser(
        'cat', parents=[multi_input_profile_parser, output_profile_parser],
        description=doc_split(cat))
    parser_cat.add_argument(
        '-x', '--prefixes', dest='prefixes', metavar='PREFIX', type=str,
        nargs='+', help='prefixes to use for the saved k-mer profile names, '
        'one per INPUT (default: profile names are assumed to be disjoint '
        'and no prefix is used)')
    parser_cat.set_defaults(func=cat)

    parser_count = subparsers.add_parser(
        'count', parents=[multi_input_parser, output_profile_parser],
        description=doc_split(count))
    parser_count.add_argument(
        '-p', '--profiles', dest='names', metavar='NAME', type=str, nargs='+',
        help='names for the created k-mer profiles, one per INPUT (default: '
        'profiles are named according to the input filenames, or numbered '
        'consecutively from 1 if no filenames are available)')
    parser_count.add_argument(
        '-k', dest='size', metavar='SIZE', type=int, default=9,
        help='k-mer size (%(type)s default: %(default)s)')
    parser_count.add_argument(
        '--by-record', '-r', dest='by_record', action='store_true',
        help='make a k-mer profile per FASTA record instead of a k-mer '
        'profile per FASTA file (profiles are named by the record names and '
        ' prefixed according to --profiles if more than one INPUT is given)')
    parser_count.set_defaults(func=count)

    parser_merge = subparsers.add_parser(
        'merge', parents=[paired_input_profile_parser, output_profile_parser],
        description=doc_split(merge))
    parser_merge.add_argument(
        '-m', dest='merger', type=str, default='sum',
        choices=metrics.mergers, help='merge function (default: %(default)s)')
    # Todo: Note in the documentation that the custom merger must be
    #   vectorized with a tip for using np.vectorize if it is not.
    parser_merge.add_argument(
        '-c', '--custom-merger', dest='custom_merger', metavar='STRING',
        type=str,
        help='custom Python merge function, specified either by an '
        'expression over the two NumPy ndarrays "left" and "right" (e.g., '
        '"np.add(left, right)"), or an importable name (e.g., '
        '"package.module.merge") that can be called with two ndarrays as '
        'arguments')
    parser_merge.set_defaults(func=merge)

    parser_balance = subparsers.add_parser(
        'balance', parents=[input_profile_parser, output_profile_parser],
        description=doc_split(balance))
    parser_balance.set_defaults(func=balance)

    parser_showbalance = subparsers.add_parser(
        'showbalance', parents=[input_profile_parser, precision_parser],
        description=doc_split(get_balance))
    parser_showbalance.set_defaults(func=get_balance, output_handle=sys.stdout)

    parser_stats = subparsers.add_parser(
        'stats', parents=[input_profile_parser, precision_parser],
        description=doc_split(get_stats))
    parser_stats.set_defaults(func=get_stats, output_handle=sys.stdout)

    parser_distr = subparsers.add_parser(
        'distr', parents=[input_profile_parser, output_parser],
        description=doc_split(distribution))
    parser_distr.set_defaults(func=distribution)

    parser_info = subparsers.add_parser(
        'info', parents=[input_profile_parser],
        description=doc_split(info))
    parser_info.set_defaults(func=info, output_handle=sys.stdout)

    parser_getcount = subparsers.add_parser(
        'getcount', parents=[input_profile_parser],
        description=doc_split(get_count))
    parser_getcount.add_argument(
        'word', metavar='WORD', type=str, help='the word in question')
    parser_getcount.set_defaults(func=get_count, output_handle=sys.stdout)

    parser_positive = subparsers.add_parser(
        'positive', parents=[paired_input_profile_parser,
                             paired_output_profile_parser],
        description=doc_split(positive))
    parser_positive.set_defaults(func=positive)

    parser_scale = subparsers.add_parser(
        'scale', parents=[paired_input_profile_parser,
                          paired_output_profile_parser, scale_parser],
        description=doc_split(scale))
    parser_scale.set_defaults(func=scale)

    parser_shrink = subparsers.add_parser(
        'shrink', parents=[input_profile_parser, output_profile_parser],
        description=doc_split(shrink))
    parser_shrink.add_argument(
        '-f', '--factor', dest='factor', metavar='INT', type=int, default=1,
        help='shrinking factor (default: %(default)s)')
    parser_shrink.set_defaults(func=shrink)

    parser_shuffle = subparsers.add_parser(
        'shuffle', parents=[input_profile_parser, output_profile_parser],
        description=doc_split(shuffle))
    parser_shuffle.set_defaults(func=shuffle)

    parser_smooth = subparsers.add_parser(
        'smooth', parents=[paired_input_profile_parser,
                           paired_output_profile_parser, smooth_parser],
        description=doc_split(smooth))
    parser_smooth.set_defaults(func=smooth)

    parser_distance = subparsers.add_parser(
        'distance', parents=[paired_input_profile_parser, dist_parser],
        description=doc_split(distance))
    parser_distance.set_defaults(func=distance, output_handle=sys.stdout)

    # Todo: I think we should just write to stdout.
    parser_matrix = subparsers.add_parser(
        'matrix', parents=[input_profile_parser, output_parser, dist_parser],
        description=doc_split(distance_matrix))
    parser_matrix.set_defaults(func=distance_matrix)

    try:
        arguments = parser.parse_args(args)
    except IOError as error:
        parser.error(error)

    try:
        arguments.func(**dict((k, v) for k, v in vars(arguments).items()
                              if k not in ('func', 'subcommand')))
    except ValueError as error:
        parser.error(error)
